import json
from datetime import date, datetime
from typing import Any, TypeVar

from pydantic import BaseModel as PydanticBaseModel

T = TypeVar("T", bound="BaseModel")


class BaseModel(PydanticBaseModel):
    @classmethod
    def from_dict(cls: type[T], data: dict[Any, Any]) -> T:
        return cls(**data)

    @classmethod
    def from_json(cls: type[T], json_data: str) -> T:
        data = json.loads(json_data)
        return cls.from_dict(data)


class Staff(BaseModel):
    id: str
    first_name: str
    last_name: str
    short_name: str
    url: str | None

    class Config:
        fields = {
            "first_name": {"alias": "firstname"},
            "last_name": {"alias": "lastname"},
            "short_name": {"alias": "shortname"},
        }


class Semester(BaseModel):
    id: str
    publish_timetable: bool
    href: str


class ExternalStaff(BaseModel):
    name: str
    url: str | None


class EventAlert(BaseModel):
    id: int
    deleted: bool
    message: str | None
    expire_date: date | None


class EventResource(BaseModel):
    name: str
    url: str | None
    type: str | None


class EventRoom(BaseModel):
    """
    Event rooms are attached to events. In addition to generic information about
    the room, they may contain metadata specific to the event in question.
    """

    id: str
    name: str
    acronym: str | None
    # if video_link is true, url is *not* a generic link for information
    # about the room, but a video meeting link which should be treated as
    # sensitive information
    url: str | None
    room_number: str
    building_id: str | None
    building_name: str | None
    building_acronym: str | None
    building_url: str | None
    campus_id: str | None

    # fields below this comment don't apply to the room in general,
    # they are related to the event

    show_for_students: bool | None
    video_link: bool | None

    # known to be included for resource bookings
    place_booking: bool | None
    places: list[int] | None
    confirmation_required: bool | None

    class Config:
        fields = {
            "name": {"alias": "roomname"},
            "acronym": {"alias": "roomacronym"},
            "url": {"alias": "roomurl"},
            "room_number": {"alias": "roomid"},
            "building_id": {"alias": "buildingid"},
            "building_name": {"alias": "buildingname"},
            "building_acronym": {"alias": "buildingacronym"},
            "building_url": {"alias": "buildingurl"},
            "campus_id": {"alias": "campusid"},
            "show_for_students": {"alias": "showforstudent"},
            "video_link": {"alias": "videolink"},
        }


class Event(BaseModel):
    id: str
    event_id: str | None
    semester_id: str
    course_id: str
    course_version: str | None
    activity_id: str | None
    activity_time_id: str | None
    week_number: int
    weekday: int | None
    # term_number may be null, in that case it may be assumed to be 1
    term_number: int | None
    start: datetime
    end: datetime
    summary: str | None
    teaching_method: str | None
    teaching_method_name: str | None
    teaching_title: str
    rooms: list[EventRoom] | None
    edit_url: str | None
    staff: list[Staff] | None
    staff_uids: list[str] | None
    staff_names: list[str] | None
    external_staff: list[ExternalStaff] | None
    title: str | None
    status: str | None
    status_plenary: bool | None
    curriculum: str | None
    serial_number: int | None
    compulsory: bool | None
    tags: list[str] | None
    party: str | None
    customer_id: str | None
    discipline: list[str] | None
    resources: list[EventResource] | None
    # details: str  # TODO: what details?

    # The following are only included for "exam events"
    exam_notes: str | None
    # Mandatory field for exams but never present for ordinary lectures
    examform_code: str | None
    examform_no: str | None
    examform_nn: str | None
    examform_en: str | None
    assessment_code: str | None
    assessment_time_year: int | None
    assessment_time_code: str | None
    alerts: list[EventAlert] | None

    class Config:
        fields = {
            "event_id": {"alias": "eventid"},
            "semester_id": {"alias": "semesterid"},
            "course_id": {"alias": "courseid"},
            "course_version": {"alias": "courseversion"},
            "activity_id": {"alias": "actid"},
            "activity_time_id": {"alias": "aid"},
            "week_number": {"alias": "weeknr"},
            "term_number": {"alias": "terminnr"},
            "start": {"alias": "dtstart"},
            "end": {"alias": "dtend"},
            "teaching_method": {"alias": "teaching-method"},
            "teaching_method_name": {"alias": "teaching-method-name"},
            "teaching_title": {"alias": "teaching-title"},
            "rooms": {"alias": "room"},
            "edit_url": {"alias": "editurl"},
            "staff": {"alias": "staffs"},
            "staff_uids": {"alias": "staff"},
            "staff_names": {"alias": "staffnames"},
            "external_staff": {"alias": "xstaff-list"},
            "curriculum": {"alias": "curr"},
            "serial_number": {"alias": "lopenr"},
            "exam_notes": {"alias": "notes"},
            "examform_code": {"alias": "examform-code"},
            "examform_no": {"alias": "examform-no"},
            "examform_nn": {"alias": "examform-nn"},
            "examform_en": {"alias": "examform-en"},
            "assessment_code": {"alias": "assessment-code"},
            "assessment_time_year": {"alias": "assessment-time-year"},
            "assessment_time_code": {"alias": "assessment-time-code"},
            "customer_id": {"alias": "customerid"},
        }


class EvuEvent(Event):
    evu_activity_id: str
    code: str
    period: str
    utype: str
    virtual_course_name: str

    class Config:
        fields = {
            "evu_activity_id": {"alias": "continuing_eduid"},
            "code": {"alias": "continuing_edu_fscode"},
            "period": {"alias": "continuing_edu_fstime"},
            "virtual_course_name": {"alias": "virtual-course-name"},
        }


class ProgramEvent(Event):
    """At least one of program/yeargroup/class is required."""

    virtual_course_name: str | None
    study_program_id: str | None
    yeargroup_id: str | None
    yeargroup_class_id: str | None

    class Config:
        fields = {
            "virtual_course_name": {"alisas": "virtual-course-name"},
            "study_program_id": {"alias": "study_programid"},
            "yeargroup_id": {"alias": "yeargroupid"},
            "yeargroup_class_id": {"alias": "yeargroup_classid"},
        }


class BaseTimetable(BaseModel):
    name: str
    version: str
    interval_start: date | None
    interval_end: date | None
    semester: str | None
    type_name: str
    last_changed: datetime | None
    timestamp: datetime
    events: list[Event]

    class Config:
        fields = {
            "interval_start": {"alias": "interval-start"},
            "interval_end": {"alias": "interval-end"},
            "type_name": {"alias": "typename"},
            "last_changed": {"alias": "lastchanged"},
        }


class CourseTimetable(BaseTimetable):
    course_id: str
    course_name: str
    title: str
    detail_type: str | None

    class Config:
        fields = {
            "course_id": {"alias": "courseid"},
            "course_name": {"alias": "coursename"},
            "detail_type": {"alias": "detail-type"},
        }


class StaffTimetable(BaseTimetable):
    staff_id: str

    class Config:
        fields = {
            "staff_id": {"alias": "staffid"},
        }


class EvuInfo(BaseModel):
    """test"""

    id: str
    code: str
    period: str
    name: str | None
    name_nn: str | None
    name_en: str | None
    department_id: str
    active: bool

    class Config:
        fields = {
            "code": {"alias": "fscode"},
            "period": {"alias": "fstime"},
            "department_id": {"alias": "departmentid"},
        }


class EvuTimetable(BaseTimetable):
    """
    Model for a EVU course timetable.

    The model for EVU courses in TP is a bit strange..
    All of the EVU courses are group together into one virtual course.
    This create problems for us, as all events will have
    the same course id.

    In FS, an EVU-course is identified by a code and a period.

        "kode": "11-RESA2702V",
        "tidsangivelse": "2021H-2022V",

    In TP, this is combined to create a activity_id of <kode>-<tidsangivelse>
    """

    activity_id: str
    events: list[EvuEvent]
    continuing_edu: EvuInfo

    class Config:
        fields = {
            "activity_id": {"alias": "activityid"},
        }


class StudentTimetable(BaseTimetable):
    student_id: str

    class Config:
        fields = {"student_id": {"alias": "studentid"}}


class BaseProgramTimetable(BaseTimetable):
    """
    This timetable type is common for all timetables retrieved via
    the program timetable endpoint.

    The ID of the timetable that was looked up is added as an additional clue
    for applications that need this context to be available in the model.
    """

    semesters: list[Semester]
    events: list[ProgramEvent]


class ProgramTimetable(BaseProgramTimetable):
    # added by this client library
    requested_program_id: str


class YeargroupTimetable(BaseProgramTimetable):
    # added by this client library
    requested_yeargroup_id: str


class ClassTimetable(BaseProgramTimetable):
    # added by this client library
    requested_class_id: str


class PersonalBookingEvent(BaseModel):
    """
    These are returned from the booking.php endpoint.

    They have various fake course IDs like "BOOKING", "BOOKING-S", "TP-RESOURCE", etc.
    """

    id: str
    activity_time_id: str | None
    course_id: str | None
    semester_id: str | None
    teaching_title: str
    teaching_method: str | None
    teaching_method_name: str | None
    summary: str | None
    start: str
    end: str
    staff_uids: list[str] | None
    staff: list[Staff] | None
    week_number: int
    customer_id: str | None
    weekday: int | None
    event_id: str | None
    edit_url: str | None
    curriculum: str | None
    resources: list[EventResource] | None
    rooms: list[EventRoom] | None

    # TODO: fields above seem to be common for most events,
    # but we cannot subclass Event because these don't contain
    # a `semesterid` field. consider creating a BaseEvent class
    # containing fields that are populated for all event types

    teaching_title_en: str | None
    teaching_title_nn: str | None
    summary_en: str | None
    summary_nn: str | None
    publish: bool | None

    # known to be included for resource bookings
    all_weeks: list[int] | None
    multiday: bool | None
    reservation_id: int | None
    contact: str | None

    class Config:
        fields = {
            "activity_time_id": {"alias": "aid"},
            "course_id": {"alias": "courseid"},
            "semester_id": {"alias": "semesterid"},
            "teaching_title": {"alias": "teaching-title"},
            "teaching_title_en": {"alias": "teaching-title_en"},
            "teaching_title_nn": {"alias": "teaching-title_nn"},
            "teaching_method": {"alias": "teaching-method"},
            "teaching_method_name": {"alias": "teaching-method-name"},
            "start": {"alias": "dtstart"},
            "end": {"alias": "dtend"},
            "week_number": {"alias": "weeknr"},
            "customer_id": {"alias": "customerid"},
            "event_id": {"alias": "eventid"},
            "edit_url": {"alias": "editurl"},
            "curriculum": {"alias": "curr"},
            "rooms": {"alias": "room"},
            "staff_uids": {"alias": "staff"},
            "staff": {"alias": "staffs"},
            "all_weeks": {"alias": "allweeks"},
            "reservation_id": {"alias": "reservationid"},
        }


class BookingsTimetable(BaseTimetable):
    """
    Personal orders

    Norwegian: Personlige bestillinger
    """

    uid: str
    events: list[PersonalBookingEvent]

    class Config:
        fields = {"uid": {"alias": "userid"}}


class Room(BaseModel):
    id: str
    internal_id: str
    name: str
    acronym: str | None
    url: str | None
    description: str | None
    type_id: str
    type_name: str
    building_id: str | None
    building_name: str | None
    building_acronym: str | None
    building_url: str | None
    audio_video_description: str
    floor: str
    capacity: str
    capacity_exam: str
    image_url: str | None
    # active_from: date
    # active_to: date  # may be empty string
    # groups = List[???]

    class Config:
        fields = {
            "internal_id": {"alias": "roomid"},
            "name": {"alias": "roomname"},
            "acronym": {"alias": "roomacronym"},
            "url": {"alias": "roomurl"},
            "type_id": {"alias": "roomtypeid"},
            "type_name": {"alias": "roomtypename"},
            "building_id": {"alias": "buildingid"},
            "building_name": {"alias": "buildingname"},
            "building_acronym": {"alias": "buildingacronym"},
            "building_url": {"alias": "buildingurl"},
            "audio_video_description": {"alias": "avtext"},
            "floor": {"alias": "floornum"},
            "capacity": {"alias": "size"},
            "capacity_exam": {"alias": "size_exam"},
            "image_url": {"alias": "roomimg_url"},
        }


class RoomData(BaseModel):
    name: str
    version: str
    timestamp: datetime
    room: Room
    equipment: list[str]

    class Config:
        fields = {
            "room": {"alias": "data"},
        }


class ListedRoom(BaseModel):
    id: str
    name: str
    type: str
    fs_room_id: str | None
    room_id: str | None
    type_id: str | None
    building_name: str
    area_name: str
    building_id: str
    area_id: str
    building_url: str | None
    room_url: str | None

    class Config:
        fields = {
            "fs_room_id": {"alias": "fsroomid"},
            "room_id": {"alias": "roomid"},
            "type_id": {"alias": "typeid"},
            "building_name": {"alias": "buildingname"},
            "area_name": {"alias": "areaname"},
            "building_id": {"alias": "buildingid"},
            "area_id": {"alias": "areaid"},
            "builing_url": {"alias": "builingurl"},
            "room_url": {"alias": "roomurl"},
        }


class Rooms(BaseModel):
    name: str
    version: str
    timestamp: datetime
    rooms: list[ListedRoom]

    class Config:
        fields = {"rooms": {"alias": "data"}}
